"""
The python source code.
"""

from SpAlgo._array import Array
from SpAlgo._matrix import Matrix
from SpAlgo._sequence import MaxSubseq
from SpAlgo._knapsack._knapsack import Knapsack

__version__ = "0.2.1"
__all__ = ['Matrix', 'MaxSubseq', 'Array', 'Knapsack']



from typing import Union


class Array:
    _inner_array = []
    _size = 0
    _all_peaks = []
    """
    :var _inner_array: 
    """
    def __init__(self, array: list[Union[int, float]]):
        self._inner_array = array
        self._size = len(self._inner_array)
        self._all_peaks = Array.findAllPeaks(self)

    def findPeak(self) -> tuple[Union[int, float], int]:
        """
        :return: a tuple of peak value and index
        """
        # if there is one element in array return it
        if self._size < 2:
            return self._inner_array[0], 0
    
        # if the first or last element of data array
        if self._inner_array[0] > self._inner_array[1]:
            return self._inner_array[0], 0
        if self._inner_array[(self._size - 1)] > self._inner_array[(self._size - 2)]:
            return self._inner_array[(self._size - 1)], (self._size - 1)
    
        # otherwise
        pointer = self._size // 2
        while True:
            current_element = self._inner_array[pointer]
            if self._inner_array[pointer + 1] > current_element:
                pointer += 1
                current_element = self._inner_array[pointer] 
    
            elif self._inner_array[pointer - 1] > current_element:
                pointer -= 1
                current_element = self._inner_array[pointer]
    
            else:
                break
        return current_element, pointer

    def findAllPeaks(self) -> list[tuple[Union[int, float]], int]:
        """
        :return: a list of tuples of All Peaks
        """
        list_of_peaks = []
        if self._inner_array[0] > self._inner_array[1]:
            list_of_peaks += [(self._inner_array[0], 0)]
        for i in range(1, self._size - 1):
            current_element = self._inner_array[i]
            if (current_element > self._inner_array[i + 1]) and (current_element > self._inner_array[i - 1]):
                list_of_peaks += [(self._inner_array[i], i)]

        if self._inner_array[self._size - 1] > self._inner_array[self._size - 2]:
            list_of_peaks += [(self._inner_array[self._size-1], self._size - 1)]

        return list_of_peaks


    
    """
        var _all_peaks defined to use in functions below
    """

    def findFirstPeak(self) -> tuple[int, int]:
        first_peak = self._all_peaks[0][0]
        return first_peak, self._inner_array.index(first_peak) 

    def findMaxPeak(self) -> tuple[int, int]:
        """
        :return the FIRST maximum peak
        """
        max_peak = max(self._all_peaks)[0]
        return max_peak, self._inner_array.index(max_peak) 

    def findMinPeak(self) -> tuple[int, int]:
        """
        :return the FIRST minimum peak
        """
        min_peak = min(self._all_peaks)[0]
        return min_peak, self._inner_array.index(min_peak)


    
        
    




from . import _pack

def _partition_(_items, low, high):
    i = low - 1
    pivot = _items[high]
    for j in range(low, high):
        if _items[j] <= pivot:
            i = i + 1
            _items[i], _items[j] = _items[j], _items[i]

    _items[i + 1], _items[high] = _items[high], _items[i + 1]
    return i + 1

def _sort_(_items, low, high):
    if len(_items) == 1:
        return _items
    if low < high:
        pi = _partition_(_items, low, high)
        _sort_(_items, low, pi - 1)
        _sort_(_items, pi + 1, high)


from typing import Union

class MaxSubseq:
    _sequence = None
    _maximum_summation = None
    start_index = -1
    end_index = -1
    """
        :var sequence: sequence of numbers which will be given at first
    """

    def __init__(self, sequence: list[Union[int, float]]):
        """
            The given sequence is a list of type int or float by the
             other word must be a numerical list.

            All list's elements can have different types:
                * float
                * integer
            :param sequence: sequence of numbers which will be given at first
        """
        self.sequence = sequence
        self.__calculate_max_longest_subsequence()

    def __calculate_max_longest_subsequence(self):
        """
          the calculate_max_longest_subsequence will calculate the maximum value of the
            longest subsequence

            type:
           :return: returns no info but will store following information->
                |---- 1. first element is the max sum of the sequence
                +-----2. the second is index of start and end of the
                         subsequence
        """
        # variables
        _current_sum = 0
        max_sum = -float('inf')
        _indexes = []
        _last_index = -1
        """
          type:
           :var _current_sum: every element of list we go through will
                be added to this variable
           :var _indexes: will hold the first index of the the subsequence
           :var max_sum: will hold the max sum of the longest subsequence
           :var _last_index: will hold the last index of subsequence
        """
        for _index in range(len(self.sequence)):
            _current_sum += self.sequence[_index]
            _indexes.append(_index)
            if _current_sum < 0:
                _current_sum = 0
                _indexes.clear()
            if _current_sum > max_sum:
                max_sum = _current_sum
                _last_index = _index

        self._maximum_summation = max_sum
        self.start_index = _indexes[0]
        self.end_index = _last_index
        return

    def maximum_sum(self):
        """
        :return: the maximum value of the subsequence
        """
        return self._maximum_summation

    def subsequence(self):
        """
        :return: will return the subsequence which is found as longest with
                 maximum summation
        """
        return self.sequence[self.start_index: self.end_index + 1]


from typing import Union
from . import _qsort
from ._pack import Pack


class Knapsack:
    """
    The Knapsack Problem is a famous Dynamic Programming Problem that falls
     in the optimization category.

    It derives its name from a scenario where, given a set of items with
    specific weights and assigned values, the goal is to maximize the value
    in a knapsack while remaining within the weight constraint.

    Each item can only be selected once, as we don’t have multiple quantities
     of any item.
    """
    _weight = None
    _value = None
    _capacity = None
    _st_capacity = None
    _pack = None
    _cur_cap = None
    _cur_val = None
    _is_crumbly = None
    _total_value = None
    _crumbled_item = None
    _infinite_item = None

    def __init__(self,
                 capacity: Union[int, float],
                 weight: list[Union[int, float]],
                 value: list[Union[int, float]],
                 is_crumbly: bool = False,
                 infinite_item: bool = False):
        """
        Initial class fields
        :param capacity: the capacity of the knapsack.
        :param weight: the array of item weights.
        :param value: the array of item values.
        :param is_crumbly: if items are crumbly so make it True. [default is False]
        :param infinite_item: if there is infinite count of items make it True. [default is False]
        """
        self._infinite_item = infinite_item
        self._value = value
        self._weight = weight
        self._capacity = capacity
        self._st_capacity = capacity
        self._is_crumbly = is_crumbly

        self._picker_()

    def getPack(self):
        """
        :return: items of knapsack
        """
        return self._pack

    def getTotalValue(self):
        """
        :return: knapsack total value
        """
        return self._total_value

    def getUsedWeight(self):
        """
        :return: used weight of the knapsack
        """
        return self._st_capacity - self._capacity

    def getWastedWeight(self):
        """
        :return: Wasted weight of the knapsack
        """
        return self._capacity

    def getCrumbledItem(self):
        """
        :return: the item which is crumbled
        the formate is:
             {
                'weight': ---,
                'value': ---,
                'original_weight': ---,
                'original_value': ---
            }
        """
        return self._crumbled_item

    def _picker_(self):
        """
        this will execute the knapsack problem
        """
        _val_pack = []
        self._pack = []
        for item in range(len(self._weight)):
            _val_pack.append(Pack(self._weight[item], self._value[item], item))

        _qsort._sort_(_val_pack, 0, len(_val_pack) - 1)
        _val_pack.reverse()

        self._total_value = 0

        if not self._infinite_item:
            for item in _val_pack:
                _cur_wt = int(item.wt)
                _cur_val = int(item.val)
                if self._capacity - _cur_wt >= 0:
                    self._itm_pack_(_cur_wt, _cur_val)
                else:
                    self._add_crumble_(_cur_wt, _cur_val)
                    break
        else:
            _inner_index = 0
            while True:
                _cur_wt = _val_pack[_inner_index].wt
                _cur_val = _val_pack[_inner_index].val

                if _val_pack[_inner_index].wt <= self._capacity:
                    self._itm_pack_(_cur_wt, _cur_val)

                if _cur_wt > self._capacity and not self._is_crumbly:
                    _inner_index += 1
                elif _cur_wt > self._capacity and self._is_crumbly:
                    self._add_crumble_(_cur_wt, _cur_val)
                    break

                if _inner_index == len(_val_pack):
                    self._add_crumble_(_cur_wt, _cur_val)
                    break

    def _add_crumble_(self, _cur_wt, _cur_val):
        """
        add crumbled item
        :param _cur_wt:
        :param _cur_val:
        :return:
        """
        if self._is_crumbly:
            fraction = self._capacity / _cur_wt
            self._total_value += _cur_val * fraction
            self._capacity = int(self._capacity - (_cur_wt * fraction))
            self._pack.append({'weight': _cur_wt * fraction, 'value': _cur_val * fraction})
            self._crumbled_item = {'weight': _cur_wt * fraction, 'value': _cur_val * fraction,
                                   'original_weight': _cur_wt, 'original_value': _cur_val}

    def _itm_pack_(self, _cur_wt, _cur_val):
        self._capacity -= _cur_wt
        self._total_value += _cur_val
        self._pack.append({'weight': _cur_wt, 'value': _cur_val})



from typing import Union
import operator


class Matrix:
    _inner_matrix = None
    _column_size = None
    _row_size = None
    _OPERATES_FLAGS = [
        {'side': 'u', 'operator': operator.sub, 'row_val': 1, 'col_val': 0},
        {'side': 'd', 'operator': operator.add, 'row_val': 1, 'col_val': 0},
        {'side': 'r', 'operator': operator.add, 'row_val': 0, 'col_val': 1},
        {'side': 'l', 'operator': operator.sub, 'row_val': 0, 'col_val': 1}
    ]

    def __init__(self, matrix: list[list[Union[int, float]]]):
        """
        :param matrix: initial the _inner_matrix with matrix given
        to the MatrixPeak.
        we will initial these variables as well:
            * :var _column_size
            * :var _row_size
        """
        self._inner_matrix = matrix
        self._row_size = len(matrix)
        self._column_size = len(matrix[0])

    def _findMaxOfColumn(self, column):
        _max = - float('inf')
        _row = None
        for row in range(self._row_size):
            if _max < self._inner_matrix[row][column]:
                _max = self._inner_matrix[row][column]
                _row = row
        return _max, _row

    def _is_edge(self, side, row=None, column=None):
        """
        :param row:
        :param column:
        :param side: is one of ['u', 'd', 'l', 'r']
        :return: boolean
        """

        if side == 'u':  # up
            if row - 1 < 0:
                return False
            return True
        elif side == 'd':  # down
            if row + 1 > self._row_size - 1:
                return False
            return True
        elif side == 'l':  # left
            if column - 1 < 0:
                return False
            return True
        elif side == 'r':  # right
            if column + 1 > self._column_size - 1:
                return False
            return True
        return False

    def findPeak(self) -> dict[str, Union[int, float]]:
        """
        This functions and allows you to find one matrix's peak
         and will return a dictionary which will have the
         following items:
            * peak
            * row
            * column
        :return: a dictionary contains these elements->
            * :var peak: contain the peak found inside matrix.
            * :var column: the column of the peak has been found.
            * :var row: the row of the peak has been found.
        """
        _middle_column = (self._column_size - 1) // 2

        _middle_column_max_element, _max_element_row = self._findMaxOfColumn(column=_middle_column)

        _current_column = _middle_column
        _current_row = _max_element_row

        _any_step = False
        while True:
            current_element = self._inner_matrix[_current_row][_current_column]
            if self._is_edge(side='d', row=_current_row):
                if self._inner_matrix[_current_row + 1][_current_column] > current_element:
                    _current_row += 1
                    _any_step = True

            if self._is_edge(side='u', row=_current_row):
                if self._inner_matrix[_current_row - 1][_current_column] > current_element:
                    _current_row -= 1
                    _any_step = True

            if self._is_edge(side='r', column=_current_column):
                if self._inner_matrix[_current_row][_current_column + 1] > current_element:
                    _current_column += 1
                    _any_step = True

            if self._is_edge(side='l', column=_current_column):
                if self._inner_matrix[_current_row][_current_column - 1] > current_element:
                    _current_column -= 1
                    _any_step = True

            if not _any_step:
                break

            _any_step = False

            current_element = self._inner_matrix[_current_row][_current_column]

        return {'peak': current_element, 'column': _current_column, 'row': _current_row}

    def _matrix_next_step(self, row, column, step=1):
        """
        this function help to go through the matrix cells.

        :param row: current row
        :param column: current column
        :param step: step forward
        :return: next row and next column
        """
        _next_row = row
        _next_column = column
        for _s_f in range(step):
            _next_column = (column + 1) % self._column_size
            if _next_column == 0:
                _next_row = (row + 1)

        return _next_row, _next_column

    def _edge_checker(self, _side, _status_matrix, current_element, _current_row, _current_column, operator_exec,
                      _sd_flags, _row_val=0, _col_val=0):
        """
        :param _side:
        :param _status_matrix:
        :param current_element:
        :param _current_row:
        :param _current_column:
        :param operator_exec:
        :param _sd_flags:
        :param _row_val:
        :param _col_val:
        :return:
        """
        _complete_side_status = 0
        _side_status = 0
        _forward_step = 1
        if self._is_edge(side=_side, row=_current_row, column=_current_column):
            _sd_flags['_complete_side_status'] += 1
            if current_element >= self._inner_matrix[operator_exec(_current_row, _row_val)][
                    operator_exec(_current_column, _col_val)]:
                _sd_flags['_side_status'] += 1
                if current_element != self._inner_matrix[operator_exec(_current_row, _row_val)][
                        operator_exec(_current_column, _col_val)]:
                    _status_matrix[operator_exec(_current_row, _row_val)][_current_column] = False
                    if _side == 'r':
                        _sd_flags['_forward_step'] = 2

    def _check_peak_pos(self, _status_matrix, current_element, _current_row, _current_column):
        """
        :param _status_matrix:
        :param current_element:
        :param _current_row:
        :param _current_column:
        :return:
        """
        _sd_flags = {'_complete_side_status': 0, '_side_status': 0, '_forward_step': 1}

        for ops_side in self._OPERATES_FLAGS:
            self._edge_checker(
                _side=ops_side['side'],
                _status_matrix=_status_matrix,
                current_element=current_element,
                _current_row=_current_row,
                _current_column=_current_column,
                operator_exec=ops_side['operator'],
                _sd_flags=_sd_flags,
                _row_val=ops_side['row_val'],
                _col_val=ops_side['col_val']
            )

        return _sd_flags['_complete_side_status'], _sd_flags['_side_status'], _sd_flags['_forward_step']

    def findAllPeak(self) -> list[dict[str, Union[int, float]]]:
        """
        These functions and allows you to find all matrix's peaks
         and will return a list of dictionary which will have the
         following item:
            * peak
            * row
            * column
        :return: list of every peak as dict
        """
        _status_matrix = []
        _col_mat_st = []
        for column in range(self._column_size):
            _col_mat_st.append(True)
        for row in range(self._row_size):
            _status_matrix.append(_col_mat_st.copy())

        _found_peaks = []

        _current_column = 0
        _current_row = 0

        while True:

            if _current_row > self._row_size - 1:
                break

            current_element = self._inner_matrix[_current_row][_current_column]
            _status = _status_matrix[_current_row][_current_column]

            if not _status:
                _current_row, _current_column = self._matrix_next_step(_current_row, _current_column)
                continue

            _complete_side_status, _side_status, _forward_step = self._check_peak_pos(
                _status_matrix=_status_matrix,
                current_element=current_element,
                _current_row=_current_row,
                _current_column=_current_column
            )

            if _side_status == _complete_side_status:
                _found_peaks.append({'peak': current_element, 'row': _current_row, 'column': _current_column})

            _current_row, _current_column = self._matrix_next_step(_current_row, _current_column, _forward_step)

        return _found_peaks



from pathlib import Path

from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()
setup(
    name='SpAlgo',
    packages=find_packages(include=['SpAlgo.*']),

    version='0.2.1',
    description='SpAlgo is a python library with special algorithms which will helps you in your education and your tasks.',
    author='Mohammad Keshavarzi & Amir Shamsi',
    url='https://github.com/Amir-Shamsi/SpAlgo',

    license='MIT',
    author_email='amirshamsi.github@gmail.com',

    github='https://github.com/mohammadkeshavarzi/SpAlgo & https://github.com/Amir-Shamsi',
    linkedin='https://linkedin.com/in/amir-shamsi & https://linkedin.com/in/mohammad-keshavarzi-1b1671218',

    install_requires=[],
    download_url='https://github.com/Amir-Shamsi/SpAlgo/archive/refs/tags/0.2.1.tar.gz',

    keywords=['SpAlgo', 'Algorithm', 'Matrix', 'Array'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



class Pack:
    def __init__(self, _weight, _value, index):
        self.wt = _weight
        self.val = _value
        self.ind = index
        self.cost = _value / _weight

    def __le__(self, other):
        return self.cost < other.cost

